import platform
import re
import subprocess
import time
from multiprocessing import Process

import pytest

from .server import run_server


# TODO: Fix and readd "memcached" (currently fails >98% of requests)
@pytest.fixture(params=("memory", "redis"))
def server(request):
    p = Process(target=run_server, args=(request.param,))
    p.start()
    time.sleep(1)
    yield
    p.terminate()
    p.join(timeout=15)


@pytest.mark.skipif(platform.python_implementation() == "PyPy", reason="Not working currently.")
def test_concurrency_error_rates(server):
    """Test with Apache benchmark tool."""

    total_requests = 1500
    # On some platforms, it's required to enlarge number of "open file descriptors"
    #  with "ulimit -n number" before doing the benchmark.
    cmd = ("ab", "-n", str(total_requests), "-c", "500", "http://127.0.0.1:8080/")
    result = subprocess.run(cmd, capture_output=True, check=True, encoding="utf-8")

    m = re.search(r"Failed requests:\s+([0-9]+)", result.stdout)
    assert m, "Missing output from ab: " + result.stdout
    failed_requests = int(m.group(1))

    m = re.search(r"Non-2xx responses:\s+([0-9]+)", result.stdout)
    non_200 = int(m.group(1)) if m else 0

    assert failed_requests / total_requests < 0.75, result.stdout
    assert non_200 / total_requests < 0.75, result.stdout



import re
from pathlib import Path

from setuptools import setup

p = Path(__file__).with_name("aiocache") / "__init__.py"
try:
    version = re.findall(r"^__version__ = \"([^']+)\"\r?$", p.read_text(), re.M)[0]
except IndexError:
    raise RuntimeError("Unable to determine version.")

readme = Path(__file__).with_name("README.rst").read_text()


setup(
    name="aiocache",
    version=version,
    author="Manuel Miranda",
    url="https://github.com/aio-libs/aiocache",
    author_email="manu.mirandad@gmail.com",
    description="multi backend asyncio cache",
    long_description=readme,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: AsyncIO",
    ],
    packages=("aiocache",),
    install_requires=None,
    extras_require={
        "redis": ["redis>=4.2.0"],
        "memcached": ["aiomcache>=0.5.2"],
        "msgpack": ["msgpack>=0.5.5"],
    },
    include_package_data=True,
)



import asyncio

import pytest

from aiocache.backends.memory import SimpleMemoryCache
from aiocache.base import _Conn
from ..utils import Keys


class TestCache:
    """
    This class ensures that all caches behave the same way and have the minimum functionality.
    To add a new cache just create the fixture for the new cache and add id as a param for the
    cache fixture
    """

    async def test_setup(self, cache):
        assert cache.namespace == "test"

    async def test_get_missing(self, cache):
        assert await cache.get(Keys.KEY) is None
        assert await cache.get(Keys.KEY, default=1) == 1

    async def test_get_existing(self, cache):
        await cache.set(Keys.KEY, "value")
        assert await cache.get(Keys.KEY) == "value"

    async def test_multi_get(self, cache):
        await cache.set(Keys.KEY, "value")
        assert await cache.multi_get([Keys.KEY, Keys.KEY_1]) == ["value", None]

    async def test_delete_missing(self, cache):
        result = await cache.delete(Keys.KEY)
        assert result == 0

    async def test_delete_existing(self, cache):
        await cache.set(Keys.KEY, "value")
        result = await cache.delete(Keys.KEY)
        assert result == 1

        value = await cache.get(Keys.KEY)
        assert value is None

    async def test_set(self, cache):
        assert await cache.set(Keys.KEY, "value") is True

    async def test_set_cancel_previous_ttl_handle(self, cache):
        await cache.set(Keys.KEY, "value", ttl=4)

        await asyncio.sleep(2.1)
        # Smaller ttl seems flaky, as if this call takes >0.5s...
        result = await cache.get(Keys.KEY)
        assert result == "value"
        await cache.set(Keys.KEY, "new_value", ttl=4)

        await asyncio.sleep(2)
        result = await cache.get(Keys.KEY)
        assert result == "new_value"

    async def test_multi_set(self, cache):
        pairs = [(Keys.KEY, "value"), [Keys.KEY_1, "random_value"]]
        assert await cache.multi_set(pairs) is True
        assert await cache.multi_get([Keys.KEY, Keys.KEY_1]) == ["value", "random_value"]

    async def test_multi_set_with_ttl(self, cache):
        pairs = [(Keys.KEY, "value"), [Keys.KEY_1, "random_value"]]
        assert await cache.multi_set(pairs, ttl=1) is True
        await asyncio.sleep(1.1)

        assert await cache.multi_get([Keys.KEY, Keys.KEY_1]) == [None, None]

    async def test_set_with_ttl(self, cache):
        await cache.set(Keys.KEY, "value", ttl=1)
        await asyncio.sleep(1.1)

        assert await cache.get(Keys.KEY) is None

    async def test_add_missing(self, cache):
        assert await cache.add(Keys.KEY, "value", ttl=1) is True

    async def test_add_existing(self, cache):
        assert await cache.set(Keys.KEY, "value") is True
        with pytest.raises(ValueError):
            await cache.add(Keys.KEY, "value")

    async def test_exists_missing(self, cache):
        assert await cache.exists(Keys.KEY) is False

    async def test_exists_existing(self, cache):
        await cache.set(Keys.KEY, "value")
        assert await cache.exists(Keys.KEY) is True

    async def test_increment_missing(self, cache):
        assert await cache.increment(Keys.KEY, delta=2) == 2
        assert await cache.increment(Keys.KEY_1, delta=-2) == -2

    async def test_increment_existing(self, cache):
        await cache.set(Keys.KEY, 2)
        assert await cache.increment(Keys.KEY, delta=2) == 4
        assert await cache.increment(Keys.KEY, delta=1) == 5
        assert await cache.increment(Keys.KEY, delta=-3) == 2

    async def test_increment_typeerror(self, cache):
        await cache.set(Keys.KEY, "value")
        with pytest.raises(TypeError):
            assert await cache.increment(Keys.KEY)

    async def test_expire_existing(self, cache):
        await cache.set(Keys.KEY, "value")
        assert await cache.expire(Keys.KEY, 1) is True
        await asyncio.sleep(1.1)
        assert await cache.exists(Keys.KEY) is False

    async def test_expire_with_0(self, cache):
        await cache.set(Keys.KEY, "value", 1)
        assert await cache.expire(Keys.KEY, 0) is True
        await asyncio.sleep(1.1)
        assert await cache.exists(Keys.KEY) is True

    async def test_expire_missing(self, cache):
        assert await cache.expire(Keys.KEY, 1) is False

    async def test_clear(self, cache):
        await cache.set(Keys.KEY, "value")
        await cache.clear()

        assert await cache.exists(Keys.KEY) is False

    async def test_close_pool_only_clears_resources(self, cache):
        await cache.set(Keys.KEY, "value")
        await cache.close()
        assert await cache.set(Keys.KEY, "value") is True
        assert await cache.get(Keys.KEY) == "value"

    async def test_single_connection(self, cache):
        async with cache.get_connection() as conn:
            assert isinstance(conn, _Conn)
            assert await conn.set(Keys.KEY, "value") is True
            assert await conn.get(Keys.KEY) == "value"


class TestMemoryCache:
    async def test_accept_explicit_args(self):
        with pytest.raises(TypeError):
            SimpleMemoryCache(random_attr="wtf")

    async def test_set_float_ttl(self, memory_cache):
        await memory_cache.set(Keys.KEY, "value", ttl=0.1)
        await asyncio.sleep(0.15)

        assert await memory_cache.get(Keys.KEY) is None

    async def test_multi_set_float_ttl(self, memory_cache):
        pairs = [(Keys.KEY, "value"), [Keys.KEY_1, "random_value"]]
        assert await memory_cache.multi_set(pairs, ttl=0.1) is True
        await asyncio.sleep(0.15)

        assert await memory_cache.multi_get([Keys.KEY, Keys.KEY_1]) == [None, None]

    async def test_raw(self, memory_cache):
        await memory_cache.raw("setdefault", "key", "value")
        assert await memory_cache.raw("get", "key") == "value"
        assert list(await memory_cache.raw("keys")) == ["key"]

    async def test_clear_with_namespace_memory(self, memory_cache):
        await memory_cache.set(Keys.KEY, "value", namespace="test")
        await memory_cache.clear(namespace="test")

        assert await memory_cache.exists(Keys.KEY, namespace="test") is False


@pytest.mark.memcached
class TestMemcachedCache:
    async def test_accept_explicit_args(self):
        from aiocache.backends.memcached import MemcachedCache

        with pytest.raises(TypeError):
            MemcachedCache(random_attr="wtf")

    async def test_set_too_long_key(self, memcached_cache):
        with pytest.raises(TypeError) as exc_info:
            await memcached_cache.set("a" * 2000, "value")
        assert str(exc_info.value).startswith("aiomcache error: invalid key")

    async def test_set_float_ttl_fails(self, memcached_cache):
        with pytest.raises(TypeError) as exc_info:
            await memcached_cache.set(Keys.KEY, "value", ttl=0.1)
        assert str(exc_info.value) == "aiomcache error: exptime not int: 0.1"

    async def test_multi_set_float_ttl(self, memcached_cache):
        with pytest.raises(TypeError) as exc_info:
            pairs = [(Keys.KEY, "value"), [Keys.KEY_1, "random_value"]]
            assert await memcached_cache.multi_set(pairs, ttl=0.1) is True
        assert str(exc_info.value) == "aiomcache error: exptime not int: 0.1"

    async def test_raw(self, memcached_cache):
        await memcached_cache.raw("set", b"key", b"value")
        assert await memcached_cache.raw("get", b"key") == "value"
        assert await memcached_cache.raw("prepend", b"key", b"super") is True
        assert await memcached_cache.raw("get", b"key") == "supervalue"

    async def test_clear_with_namespace_memcached(self, memcached_cache):
        await memcached_cache.set(Keys.KEY, "value", namespace="test")

        with pytest.raises(ValueError):
            await memcached_cache.clear(namespace="test")

        assert await memcached_cache.exists(Keys.KEY, namespace="test") is True

    async def test_close(self, memcached_cache):
        await memcached_cache.set(Keys.KEY, "value")
        await memcached_cache._close()
        assert memcached_cache.client._pool._pool.qsize() == 0


@pytest.mark.redis
class TestRedisCache:
    async def test_accept_explicit_args(self):
        from aiocache.backends.redis import RedisCache

        with pytest.raises(TypeError):
            RedisCache(random_attr="wtf")

    async def test_float_ttl(self, redis_cache):
        await redis_cache.set(Keys.KEY, "value", ttl=0.1)
        await asyncio.sleep(0.15)

        assert await redis_cache.get(Keys.KEY) is None

    async def test_multi_set_float_ttl(self, redis_cache):
        pairs = [(Keys.KEY, "value"), [Keys.KEY_1, "random_value"]]
        assert await redis_cache.multi_set(pairs, ttl=0.1) is True
        await asyncio.sleep(0.15)

        assert await redis_cache.multi_get([Keys.KEY, Keys.KEY_1]) == [None, None]

    async def test_raw(self, redis_cache):
        await redis_cache.raw("set", "key", "value")
        assert await redis_cache.raw("get", "key") == "value"
        assert await redis_cache.raw("keys", "k*") == ["key"]
        # .raw() doesn't build key with namespace prefix, clear it manually
        await redis_cache.raw("delete", "key")

    async def test_clear_with_namespace_redis(self, redis_cache):
        await redis_cache.set(Keys.KEY, "value", namespace="test")
        await redis_cache.clear(namespace="test")

        assert await redis_cache.exists(Keys.KEY, namespace="test") is False

    async def test_close(self, redis_cache):
        await redis_cache.set(Keys.KEY, "value")
        await redis_cache._close()



import asyncio

import pytest

from aiocache import Cache, caches
from ..utils import KEY_LOCK, Keys


@pytest.fixture(autouse=True)
def reset_caches():
    caches._caches = {}
    caches.set_config(
        {
            "default": {
                "cache": "aiocache.SimpleMemoryCache",
                "serializer": {"class": "aiocache.serializers.NullSerializer"},
            }
        }
    )


@pytest.fixture
async def redis_cache():
    async with Cache(Cache.REDIS, namespace="test") as cache:
        yield cache
        await asyncio.gather(*(cache.delete(k) for k in (*Keys, KEY_LOCK)))


@pytest.fixture
async def memory_cache():
    async with Cache(namespace="test") as cache:
        yield cache
        await asyncio.gather(*(cache.delete(k) for k in (*Keys, KEY_LOCK)))


@pytest.fixture
async def memcached_cache():
    async with Cache(Cache.MEMCACHED, namespace="test") as cache:
        yield cache
        await asyncio.gather(*(cache.delete(k) for k in (*Keys, KEY_LOCK)))


@pytest.fixture(
    params=(
        pytest.param("redis_cache", marks=pytest.mark.redis),
        "memory_cache",
        pytest.param("memcached_cache", marks=pytest.mark.memcached),
    ))
def cache(request):
    return request.getfixturevalue(request.param)



import asyncio
import inspect
import random
import sys
from unittest.mock import ANY, create_autospec, patch

import pytest

from aiocache import cached, cached_stampede, multi_cached
from aiocache.backends.memory import SimpleMemoryCache
from aiocache.base import BaseCache, SENTINEL
from aiocache.decorators import _get_args_dict
from aiocache.lock import RedLock


async def stub(*args, value=None, seconds=0, **kwargs):
    await asyncio.sleep(seconds)
    if value:
        return str(value)
    return str(random.randint(1, 50))


class TestCached:
    @pytest.fixture
    def decorator(self, mock_cache):
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):
            yield cached()

    @pytest.fixture
    def decorator_call(self, decorator):
        d = decorator(stub)
        yield d

    @pytest.fixture(autouse=True)
    def spy_stub(self, mocker):
        module = sys.modules[globals()["__name__"]]
        mocker.spy(module, "stub")

    def test_init(self):
        c = cached(
            ttl=1,
            key="key",
            key_builder="fn",
            cache=SimpleMemoryCache,
            plugins=None,
            alias=None,
            noself=False,
            namespace="test",
            unused_kwarg="unused",
        )

        assert c.ttl == 1
        assert c.key == "key"
        assert c.key_builder == "fn"
        assert c.cache is None
        assert c._cache == SimpleMemoryCache
        assert c._serializer is None
        assert c._namespace == "test"
        assert c._kwargs == {"unused_kwarg": "unused"}

    def test_fails_at_instantiation(self):
        with pytest.raises(TypeError):

            @cached(wrong_param=1)
            async def fn() -> None:
                """Dummy function."""

    def test_alias_takes_precedence(self, mock_cache):
        with patch(
            "aiocache.decorators.caches.get", autospec=True, return_value=mock_cache
        ) as mock_get:
            c = cached(alias="default", cache=SimpleMemoryCache, namespace="test")
            c(stub)

            mock_get.assert_called_with("default")
            assert c.cache is mock_cache

    def test_get_cache_key_with_key(self, decorator):
        decorator.key = "key"
        decorator.key_builder = "fn"
        assert decorator.get_cache_key(stub, (1, 2), {"a": 1, "b": 2}) == "key"

    def test_get_cache_key_without_key_and_attr(self, decorator):
        assert (
            decorator.get_cache_key(stub, (1, 2), {"a": 1, "b": 2})
            == "stub(1, 2)[('a', 1), ('b', 2)]"
        )

    def test_get_cache_key_without_key_and_attr_noself(self, decorator):
        decorator.noself = True
        assert (
            decorator.get_cache_key(stub, ("self", 1, 2), {"a": 1, "b": 2})
            == "stub(1, 2)[('a', 1), ('b', 2)]"
        )

    def test_get_cache_key_with_key_builder(self, decorator):
        decorator.key_builder = lambda *args, **kwargs: kwargs["market"].upper()
        assert decorator.get_cache_key(stub, (), {"market": "es"}) == "ES"

    async def test_calls_get_and_returns(self, decorator, decorator_call):
        decorator.cache.get.return_value = 1

        await decorator_call()

        decorator.cache.get.assert_called_with("stub()[]")
        assert decorator.cache.set.call_count == 0
        assert stub.call_count == 0

    async def test_cache_read_disabled(self, decorator, decorator_call):
        await decorator_call(cache_read=False)

        assert decorator.cache.get.call_count == 0
        assert decorator.cache.set.call_count == 1
        assert stub.call_count == 1

    async def test_cache_write_disabled(self, decorator, decorator_call):
        decorator.cache.get.return_value = None

        await decorator_call(cache_write=False)

        assert decorator.cache.get.call_count == 1
        assert decorator.cache.set.call_count == 0
        assert stub.call_count == 1

    async def test_disable_params_not_propagated(self, decorator, decorator_call):
        decorator.cache.get.return_value = None

        await decorator_call(cache_read=False, cache_write=False)

        stub.assert_called_once_with()

    async def test_get_from_cache_returns(self, decorator, decorator_call):
        decorator.cache.get.return_value = 1
        assert await decorator.get_from_cache("key") == 1

    async def test_get_from_cache_exception(self, decorator, decorator_call):
        decorator.cache.get.side_effect = Exception
        assert await decorator.get_from_cache("key") is None

    async def test_get_from_cache_none(self, decorator, decorator_call):
        decorator.cache.get.return_value = None
        assert await decorator.get_from_cache("key") is None

    async def test_calls_fn_set_when_get_none(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "get_from_cache")
        mocker.spy(decorator, "set_in_cache")
        decorator.cache.get.return_value = None

        await decorator_call(value="value")

        assert decorator.get_from_cache.call_count == 1
        decorator.set_in_cache.assert_called_with("stub()[('value', 'value')]", "value")
        stub.assert_called_once_with(value="value")

    async def test_calls_fn_raises_exception(self, decorator, decorator_call):
        decorator.cache.get.return_value = None
        stub.side_effect = Exception()
        with pytest.raises(Exception):
            assert await decorator_call()

    async def test_cache_write_waits_for_future(self, decorator, decorator_call):
        with patch.object(decorator, "get_from_cache", autospec=True, return_value=None) as m:
            await decorator_call()

            m.assert_awaited()

    async def test_cache_write_doesnt_wait_for_future(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "set_in_cache")
        with patch.object(decorator, "get_from_cache", autospec=True, return_value=None):
            with patch("aiocache.decorators.asyncio.ensure_future", autospec=True):
                await decorator_call(aiocache_wait_for_write=False, value="value")

        decorator.set_in_cache.assert_not_awaited()
        decorator.set_in_cache.assert_called_once_with("stub()[('value', 'value')]", "value")

    async def test_set_calls_set(self, decorator, decorator_call):
        await decorator.set_in_cache("key", "value")
        decorator.cache.set.assert_called_with("key", "value", ttl=SENTINEL)

    async def test_set_calls_set_ttl(self, decorator, decorator_call):
        decorator.ttl = 10
        await decorator.set_in_cache("key", "value")
        decorator.cache.set.assert_called_with("key", "value", ttl=decorator.ttl)

    async def test_set_catches_exception(self, decorator, decorator_call):
        decorator.cache.set.side_effect = Exception
        assert await decorator.set_in_cache("key", "value") is None

    async def test_decorate(self, mock_cache):
        mock_cache.get.return_value = None
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):

            @cached()
            async def fn(n):
                return n

            assert await fn(1) == 1
            assert await fn(2) == 2
            assert fn.cache == mock_cache

    async def test_keeps_signature(self, mock_cache):
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):

            @cached()
            async def what(self, a, b):
                """Dummy function."""

            assert what.__name__ == "what"
            assert str(inspect.signature(what)) == "(self, a, b)"
            assert inspect.getfullargspec(what.__wrapped__).args == ["self", "a", "b"]

    async def test_reuses_cache_instance(self):
        with patch("aiocache.decorators._get_cache", autospec=True) as get_c:
            cache = create_autospec(BaseCache, instance=True)
            get_c.side_effect = [cache, None]

            @cached()
            async def what():
                """Dummy function."""

            await what()
            await what()

            assert get_c.call_count == 1
            assert cache.get.call_count == 2

    async def test_cache_per_function(self):
        @cached()
        async def foo():
            """First function."""

        @cached()
        async def bar():
            """Second function."""

        assert foo.cache != bar.cache


class TestCachedStampede:
    @pytest.fixture
    def decorator(self, mock_cache):
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):
            yield cached_stampede()

    @pytest.fixture
    def decorator_call(self, decorator):
        yield decorator(stub)

    @pytest.fixture(autouse=True)
    def spy_stub(self, mocker):
        module = sys.modules[globals()["__name__"]]
        mocker.spy(module, "stub")

    def test_inheritance(self):
        assert isinstance(cached_stampede(), cached)

    def test_init(self):
        c = cached_stampede(
            lease=3,
            ttl=1,
            key="key",
            key_builder="fn",
            cache=SimpleMemoryCache,
            plugins=None,
            alias=None,
            noself=False,
            namespace="test",
            unused_kwarg="unused",
        )

        assert c.ttl == 1
        assert c.key == "key"
        assert c.key_builder == "fn"
        assert c.cache is None
        assert c._cache == SimpleMemoryCache
        assert c._serializer is None
        assert c.lease == 3
        assert c._namespace == "test"
        assert c._kwargs == {"unused_kwarg": "unused"}

    async def test_calls_get_and_returns(self, decorator, decorator_call):
        decorator.cache.get.return_value = 1

        await decorator_call()

        decorator.cache.get.assert_called_with("stub()[]")
        assert decorator.cache.set.call_count == 0
        assert stub.call_count == 0

    async def test_calls_fn_raises_exception(self, decorator, decorator_call):
        decorator.cache.get.return_value = None
        stub.side_effect = Exception()
        with pytest.raises(Exception):
            assert await decorator_call()

    async def test_calls_redlock(self, decorator, decorator_call):
        decorator.cache.get.return_value = None
        lock = create_autospec(RedLock, instance=True)

        with patch("aiocache.decorators.RedLock", autospec=True, return_value=lock):
            await decorator_call(value="value")

            assert decorator.cache.get.call_count == 2
            assert lock.__aenter__.call_count == 1
            assert lock.__aexit__.call_count == 1
            decorator.cache.set.assert_called_with(
                "stub()[('value', 'value')]", "value", ttl=SENTINEL
            )
            stub.assert_called_once_with(value="value")

    async def test_calls_locked_client(self, decorator, decorator_call):
        decorator.cache.get.side_effect = [None, None, None, "value"]
        decorator.cache._add.side_effect = [True, ValueError]
        lock1 = create_autospec(RedLock, instance=True)
        lock2 = create_autospec(RedLock, instance=True)

        with patch("aiocache.decorators.RedLock", autospec=True, side_effect=[lock1, lock2]):
            await asyncio.gather(decorator_call(value="value"), decorator_call(value="value"))

            assert decorator.cache.get.call_count == 4
            assert lock1.__aenter__.call_count == 1
            assert lock1.__aexit__.call_count == 1
            assert lock2.__aenter__.call_count == 1
            assert lock2.__aexit__.call_count == 1
            decorator.cache.set.assert_called_with(
                "stub()[('value', 'value')]", "value", ttl=SENTINEL
            )
            assert stub.call_count == 1


async def stub_dict(*args, keys=None, **kwargs):
    values = {"a": random.randint(1, 50), "b": random.randint(1, 50), "c": random.randint(1, 50)}
    return {k: values.get(k) for k in keys}


class TestMultiCached:
    @pytest.fixture
    def decorator(self, mock_cache):
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):
            yield multi_cached(keys_from_attr="keys")

    @pytest.fixture
    def decorator_call(self, decorator):
        d = decorator(stub_dict)
        decorator._conn = decorator.cache.get_connection()
        yield d

    @pytest.fixture(autouse=True)
    def spy_stub_dict(self, mocker):
        module = sys.modules[globals()["__name__"]]
        mocker.spy(module, "stub_dict")

    def test_init(self):
        mc = multi_cached(
            keys_from_attr="keys",
            key_builder=None,
            ttl=1,
            cache=SimpleMemoryCache,
            plugins=None,
            alias=None,
            namespace="test",
            unused_kwarg="unused",
        )

        def f():
            """Dummy function. Not called."""

        assert mc.ttl == 1
        assert mc.key_builder("key", f) == "key"
        assert mc.keys_from_attr == "keys"
        assert mc.cache is None
        assert mc._cache == SimpleMemoryCache
        assert mc._serializer is None
        assert mc._namespace == "test"
        assert mc._kwargs == {"unused_kwarg": "unused"}

    def test_fails_at_instantiation(self):
        with pytest.raises(TypeError):

            @multi_cached(wrong_param=1)
            async def fn() -> None:
                """Dummy function."""

    def test_alias_takes_precedence(self, mock_cache):
        with patch(
            "aiocache.decorators.caches.get", autospec=True, return_value=mock_cache
        ) as mock_get:
            mc = multi_cached(
                keys_from_attr="keys", alias="default", cache=SimpleMemoryCache, namespace="test"
            )
            mc(stub_dict)

            mock_get.assert_called_with("default")
            assert mc.cache is mock_cache

    def test_get_cache_keys(self, decorator):
        keys = decorator.get_cache_keys(stub_dict, (), {"keys": ["a", "b"]})
        assert keys == (["a", "b"], [], -1)

    def test_get_cache_keys_empty_list(self, decorator):
        assert decorator.get_cache_keys(stub_dict, (), {"keys": []}) == ([], [], -1)

    def test_get_cache_keys_missing_kwarg(self, decorator):
        assert decorator.get_cache_keys(stub_dict, (), {}) == ([], [], -1)

    def test_get_cache_keys_arg_key_from_attr(self, decorator):
        def fake(keys, a=1, b=2):
            """Dummy function."""

        assert decorator.get_cache_keys(fake, (["a"]), {}) == (["a"], [["a"]], 0)

    def test_get_cache_keys_with_none(self, decorator):
        assert decorator.get_cache_keys(stub_dict, (), {"keys": None}) == ([], [], -1)

    def test_get_cache_keys_with_key_builder(self, decorator):
        decorator.key_builder = lambda key, *args, **kwargs: kwargs["market"] + "_" + key.upper()
        assert decorator.get_cache_keys(stub_dict, (), {"keys": ["a", "b"], "market": "ES"}) == (
            ["ES_A", "ES_B"],
            [],
            -1,
        )

    async def test_get_from_cache(self, decorator, decorator_call):
        decorator.cache.multi_get.return_value = [1, 2, 3]

        assert await decorator.get_from_cache("a", "b", "c") == [1, 2, 3]
        decorator.cache.multi_get.assert_called_with(("a", "b", "c"))

    async def test_get_from_cache_no_keys(self, decorator, decorator_call):
        assert await decorator.get_from_cache() == []
        assert decorator.cache.multi_get.call_count == 0

    async def test_get_from_cache_exception(self, decorator, decorator_call):
        decorator.cache.multi_get.side_effect = Exception

        assert await decorator.get_from_cache("a", "b", "c") == [None, None, None]
        decorator.cache.multi_get.assert_called_with(("a", "b", "c"))

    async def test_get_from_cache_conn(self, decorator, decorator_call):
        decorator.cache.multi_get.return_value = [1, 2, 3]

        assert await decorator.get_from_cache("a", "b", "c") == [1, 2, 3]
        decorator.cache.multi_get.assert_called_with(("a", "b", "c"))

    async def test_calls_no_keys(self, decorator, decorator_call):
        await decorator_call(keys=[])
        assert decorator.cache.multi_get.call_count == 0
        assert stub_dict.call_count == 1

    async def test_returns_from_multi_set(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "get_from_cache")
        mocker.spy(decorator, "set_in_cache")
        decorator.cache.multi_get.return_value = [1, 2]

        assert await decorator_call(1, keys=["a", "b"]) == {"a": 1, "b": 2}
        decorator.get_from_cache.assert_called_once_with("a", "b")
        assert decorator.set_in_cache.call_count == 0
        assert stub_dict.call_count == 0

    async def test_calls_fn_multi_set_when_multi_get_none(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "get_from_cache")
        mocker.spy(decorator, "set_in_cache")
        decorator.cache.multi_get.return_value = [None, None]

        ret = await decorator_call(1, keys=["a", "b"], value="value")

        decorator.get_from_cache.assert_called_once_with("a", "b")
        decorator.set_in_cache.assert_called_with(ret, stub_dict, ANY, ANY)
        stub_dict.assert_called_once_with(1, keys=["a", "b"], value="value")

    async def test_cache_write_waits_for_future(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "set_in_cache")
        with patch.object(decorator, "get_from_cache", autospec=True, return_value=[None, None]):
            await decorator_call(1, keys=["a", "b"], value="value")

            decorator.set_in_cache.assert_awaited()

    async def test_cache_write_doesnt_wait_for_future(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "set_in_cache")
        with patch.object(decorator, "get_from_cache", autospec=True, return_value=[None, None]):
            with patch("aiocache.decorators.asyncio.ensure_future", autospec=True):
                await decorator_call(1, keys=["a", "b"], value="value",
                                     aiocache_wait_for_write=False)

        decorator.set_in_cache.assert_not_awaited()
        decorator.set_in_cache.assert_called_once_with({"a": ANY, "b": ANY}, stub_dict, ANY, ANY)

    async def test_calls_fn_with_only_missing_keys(self, mocker, decorator, decorator_call):
        mocker.spy(decorator, "set_in_cache")
        decorator.cache.multi_get.return_value = [1, None]

        assert await decorator_call(1, keys=["a", "b"], value="value") == {"a": ANY, "b": ANY}

        decorator.set_in_cache.assert_called_once_with({"a": ANY, "b": ANY}, stub_dict, ANY, ANY)
        stub_dict.assert_called_once_with(1, keys=["b"], value="value")

    async def test_calls_fn_raises_exception(self, decorator, decorator_call):
        decorator.cache.multi_get.return_value = [None]
        stub_dict.side_effect = Exception()
        with pytest.raises(Exception):
            assert await decorator_call(keys=[])

    async def test_cache_read_disabled(self, decorator, decorator_call):
        await decorator_call(1, keys=["a", "b"], cache_read=False)

        assert decorator.cache.multi_get.call_count == 0
        assert decorator.cache.multi_set.call_count == 1
        assert stub_dict.call_count == 1

    async def test_cache_write_disabled(self, decorator, decorator_call):
        decorator.cache.multi_get.return_value = [None, None]

        await decorator_call(1, keys=["a", "b"], cache_write=False)

        assert decorator.cache.multi_get.call_count == 1
        assert decorator.cache.multi_set.call_count == 0
        assert stub_dict.call_count == 1

    async def test_disable_params_not_propagated(self, decorator, decorator_call):
        decorator.cache.multi_get.return_value = [None, None]

        await decorator_call(1, keys=["a", "b"], cache_read=False, cache_write=False)

        stub_dict.assert_called_once_with(1, keys=["a", "b"])

    async def test_set_in_cache(self, decorator, decorator_call):
        await decorator.set_in_cache({"a": 1, "b": 2}, stub_dict, (), {})

        call_args = decorator.cache.multi_set.call_args[0][0]
        assert ("a", 1) in call_args
        assert ("b", 2) in call_args
        assert decorator.cache.multi_set.call_args[1]["ttl"] is SENTINEL

    async def test_set_in_cache_with_ttl(self, decorator, decorator_call):
        decorator.ttl = 10
        await decorator.set_in_cache({"a": 1, "b": 2}, stub_dict, (), {})

        assert decorator.cache.multi_set.call_args[1]["ttl"] == decorator.ttl

    async def test_set_in_cache_exception(self, decorator, decorator_call):
        decorator.cache.multi_set.side_effect = Exception

        assert await decorator.set_in_cache({"a": 1, "b": 2}, stub_dict, (), {}) is None

    async def test_decorate(self, mock_cache):
        mock_cache.multi_get.return_value = [None]
        with patch("aiocache.decorators._get_cache", autospec=True, return_value=mock_cache):

            @multi_cached(keys_from_attr="keys")
            async def fn(keys=None):
                return {"test": 1}

            assert await fn(keys=["test"]) == {"test": 1}
            assert await fn(["test"]) == {"test": 1}
            assert fn.cache == mock_cache

    async def test_keeps_signature(self):
        @multi_cached(keys_from_attr="keys")
        async def what(self, keys=None, what=1):
            """Dummy function."""

        assert what.__name__ == "what"
        assert str(inspect.signature(what)) == "(self, keys=None, what=1)"
        assert inspect.getfullargspec(what.__wrapped__).args == ["self", "keys", "what"]

    async def test_reuses_cache_instance(self):
        with patch("aiocache.decorators._get_cache", autospec=True) as get_c:
            cache = create_autospec(BaseCache, instance=True)
            cache.multi_get.return_value = [None]
            get_c.side_effect = [cache, None]

            @multi_cached("keys")
            async def what(keys=None):
                return {}

            await what(keys=["a"])
            await what(keys=["a"])

            assert get_c.call_count == 1
            assert cache.multi_get.call_count == 2

    async def test_cache_per_function(self):
        @multi_cached("keys")
        async def foo():
            """First function."""

        @multi_cached("keys")
        async def bar():
            """Second function."""

        assert foo.cache != bar.cache


def test_get_args_dict():
    def fn(a, b, *args, keys=None, **kwargs):
        """Dummy function."""

    args_dict = _get_args_dict(fn, ("a", "b", "c", "d"), {"what": "what"})
    assert args_dict == {"a": "a", "b": "b", "keys": None, "what": "what"}



import asyncio
import random
from unittest import mock

import pytest

from aiocache import cached, cached_stampede, multi_cached
from aiocache.base import _ensure_key
from ..utils import Keys


async def return_dict(keys=None):
    ret = {}
    for value, key in enumerate(keys or [Keys.KEY, Keys.KEY_1]):
        ret[key] = str(value)
    return ret


async def stub(arg: float, seconds: int = 0) -> str:
    await asyncio.sleep(seconds)
    return str(random.randint(1, 50))


class TestCached:
    @pytest.fixture(autouse=True)
    def default_cache(self, mocker, cache):
        mocker.patch("aiocache.decorators._get_cache", autospec=True, return_value=cache)

    async def test_cached_ttl(self, cache):
        @cached(ttl=2, key=Keys.KEY)
        async def fn():
            return str(random.randint(1, 50))

        resp1 = await fn()
        resp2 = await fn()

        assert await cache.get(Keys.KEY) == resp1 == resp2
        await asyncio.sleep(2.1)
        assert await cache.get(Keys.KEY) is None

    async def test_cached_key_builder(self, cache):
        def build_key(f, self, a, b):
            return "{}_{}_{}_{}".format(self, f.__name__, a, b)

        @cached(key_builder=build_key)
        async def fn(self, a, b=2):
            return "1"

        await fn("self", 1, 3)
        assert await cache.exists(build_key(fn, "self", 1, 3)) is True

    @pytest.mark.parametrize("decorator", (cached, cached_stampede))
    async def test_cached_skip_cache_func(self, cache, decorator):
        @decorator(skip_cache_func=lambda r: r is None)
        async def sk_func(x):
            return x if x > 0 else None

        arg = 1
        res = await sk_func(arg)
        assert res

        key = decorator().get_cache_key(sk_func, args=(1,), kwargs={})

        assert key
        assert await cache.exists(key)
        assert await cache.get(key) == res

        arg = -1

        await sk_func(arg)

        key = decorator().get_cache_key(sk_func, args=(-1,), kwargs={})

        assert key
        assert not await cache.exists(key)

    async def test_cached_without_namespace(self, cache):
        """Default cache key is created when no namespace is provided"""
        @cached(namespace=None)
        async def fn():
            return "1"

        await fn()
        decorator = cached(namespace=None)
        key = decorator.get_cache_key(fn, args=(), kwargs={})
        assert await cache.exists(key, namespace=None) is True

    async def test_cached_with_namespace(self, cache):
        """Cache key is prefixed with provided namespace"""
        key_prefix = "test"

        @cached(namespace=key_prefix)
        async def ns_fn():
            return "1"

        await ns_fn()
        decorator = cached(namespace=key_prefix)
        key = decorator.get_cache_key(ns_fn, args=(), kwargs={})
        assert await cache.exists(key, namespace=key_prefix) is True


class TestCachedStampede:
    @pytest.fixture(autouse=True)
    def default_cache(self, mocker, cache):
        mocker.patch("aiocache.decorators._get_cache", autospec=True, return_value=cache)

    async def test_cached_stampede(self, mocker, cache):
        mocker.spy(cache, "get")
        mocker.spy(cache, "set")
        decorator = cached_stampede(ttl=10, lease=3)

        await asyncio.gather(decorator(stub)(0.5), decorator(stub)(0.5))

        cache.get.assert_called_with("tests.acceptance.test_decoratorsstub(0.5,)[]")
        assert cache.get.call_count == 4
        cache.set.assert_called_with("tests.acceptance.test_decoratorsstub(0.5,)[]",
                                     mock.ANY, ttl=10)
        assert cache.set.call_count == 1, cache.set.call_args_list

    async def test_locking_dogpile_lease_expiration(self, mocker, cache):
        mocker.spy(cache, "get")
        mocker.spy(cache, "set")
        decorator = cached_stampede(ttl=10, lease=3)

        await asyncio.gather(
            decorator(stub)(1, seconds=1),
            decorator(stub)(1, seconds=2),
            decorator(stub)(1, seconds=3),
        )

        assert cache.get.call_count == 6
        assert cache.set.call_count == 3

    async def test_locking_dogpile_task_cancellation(self, cache):
        @cached_stampede()
        async def cancel_task():
            raise asyncio.CancelledError()

        with pytest.raises(asyncio.CancelledError):
            await cancel_task()


class TestMultiCachedDecorator:
    @pytest.fixture(autouse=True)
    def default_cache(self, mocker, cache):
        mocker.patch("aiocache.decorators._get_cache", autospec=True, return_value=cache)

    async def test_multi_cached(self, cache):
        multi_cached_decorator = multi_cached("keys")

        default_keys = {Keys.KEY, Keys.KEY_1}
        await multi_cached_decorator(return_dict)(keys=default_keys)

        for key in default_keys:
            assert await cache.get(key) is not None

    async def test_keys_without_kwarg(self, cache):
        @multi_cached("keys")
        async def fn(keys):
            return {Keys.KEY: 1}

        await fn([Keys.KEY])
        assert await cache.exists(Keys.KEY) is True

    async def test_multi_cached_key_builder(self, cache):
        def build_key(key, f, self, keys, market="ES"):
            return "{}_{}_{}".format(f.__name__, _ensure_key(key), market)

        @multi_cached(keys_from_attr="keys", key_builder=build_key)
        async def fn(self, keys, market="ES"):
            return {Keys.KEY: 1, Keys.KEY_1: 2}

        await fn("self", keys=[Keys.KEY, Keys.KEY_1])
        assert await cache.exists("fn_" + _ensure_key(Keys.KEY) + "_ES") is True
        assert await cache.exists("fn_" + _ensure_key(Keys.KEY_1) + "_ES") is True

    async def test_multi_cached_skip_keys(self, cache):
        @multi_cached(keys_from_attr="keys", skip_cache_func=lambda _, v: v is None)
        async def multi_sk_fn(keys, values):
            return {k: v for k, v in zip(keys, values)}

        res = await multi_sk_fn(keys=[Keys.KEY, Keys.KEY_1], values=[42, None])
        assert res
        assert Keys.KEY in res and Keys.KEY_1 in res

        assert await cache.exists(Keys.KEY)
        assert await cache.get(Keys.KEY) == res[Keys.KEY]
        assert not await cache.exists(Keys.KEY_1)

    async def test_fn_with_args(self, cache):
        @multi_cached("keys")
        async def fn(keys, *args):
            assert len(args) == 1
            return {Keys.KEY: 1}

        await fn([Keys.KEY], "arg")
        assert await cache.exists(Keys.KEY) is True

    async def test_double_decorator(self, cache):
        def dummy_d(fn):
            async def wrapper(*args, **kwargs):
                await fn(*args, **kwargs)

            return wrapper

        @dummy_d
        @multi_cached("keys")
        async def fn(keys):
            return {Keys.KEY: 1}

        await fn([Keys.KEY])
        assert await cache.exists(Keys.KEY) is True



import asyncio
from unittest.mock import Mock, patch

import pytest

from aiocache.lock import OptimisticLock, OptimisticLockError, RedLock
from ..utils import KEY_LOCK, Keys


class TestRedLock:
    @pytest.fixture
    def lock(self, mock_base_cache):
        RedLock._EVENTS = {}
        yield RedLock(mock_base_cache, Keys.KEY, 20)

    async def test_acquire(self, mock_base_cache, lock):
        await lock._acquire()
        mock_base_cache._add.assert_called_with(KEY_LOCK, lock._value, ttl=20)
        assert lock._EVENTS[KEY_LOCK].is_set() is False

    async def test_release(self, mock_base_cache, lock):
        mock_base_cache._redlock_release.return_value = True
        await lock._acquire()
        await lock._release()
        mock_base_cache._redlock_release.assert_called_with(KEY_LOCK, lock._value)
        assert KEY_LOCK not in lock._EVENTS

    async def test_release_no_acquire(self, mock_base_cache, lock):
        mock_base_cache._redlock_release.return_value = False
        assert KEY_LOCK not in lock._EVENTS
        await lock._release()
        assert KEY_LOCK not in lock._EVENTS

    async def test_context_manager(self, mock_base_cache, lock):
        async with lock:
            pass
        mock_base_cache._add.assert_called_with(KEY_LOCK, lock._value, ttl=20)
        mock_base_cache._redlock_release.assert_called_with(KEY_LOCK, lock._value)

    async def test_raises_exceptions(self, mock_base_cache, lock):
        mock_base_cache._redlock_release.return_value = True
        with pytest.raises(ValueError):
            async with lock:
                raise ValueError

    async def test_acquire_block_timeouts(self, mock_base_cache, lock):
        await lock._acquire()

        # Mock .wait() to avoid unawaited coroutine warning.
        with patch.object(RedLock._EVENTS[lock.key], "wait", Mock(spec_set=())):
            with patch("asyncio.wait_for", autospec=True, side_effect=asyncio.TimeoutError):
                mock_base_cache._add.side_effect = ValueError
                result = await lock._acquire()
                assert result is None

    async def test_wait_for_release_no_acquire(self, mock_base_cache, lock):
        mock_base_cache._add.side_effect = ValueError
        assert await lock._acquire() is None

    async def test_multiple_locks_lock(self, mock_base_cache, lock):
        lock_1 = RedLock(mock_base_cache, Keys.KEY, 20)
        lock_2 = RedLock(mock_base_cache, Keys.KEY, 20)
        mock_base_cache._add.side_effect = [True, ValueError(), ValueError()]
        await lock._acquire()
        event = lock._EVENTS[KEY_LOCK]

        assert KEY_LOCK in lock._EVENTS
        assert KEY_LOCK in lock_1._EVENTS
        assert KEY_LOCK in lock_2._EVENTS
        assert not event.is_set()

        await asyncio.gather(lock_1._acquire(), lock._release(), lock_2._acquire())

        assert KEY_LOCK not in lock._EVENTS
        assert KEY_LOCK not in lock_1._EVENTS
        assert KEY_LOCK not in lock_2._EVENTS
        assert event.is_set()


class TestOptimisticLock:
    @pytest.fixture
    def lock(self, mock_base_cache):
        yield OptimisticLock(mock_base_cache, Keys.KEY)

    def test_init(self, mock_base_cache, lock):
        assert lock.client == mock_base_cache
        assert lock._token is None
        assert lock.key == Keys.KEY
        assert lock.ns_key == mock_base_cache._build_key(Keys.KEY)

    async def test_aenter_returns_lock(self, lock):
        assert await lock.__aenter__() is lock

    async def test_aexit_not_crashing(self, lock):
        async with lock:
            pass

    async def test_acquire_calls_get(self, lock):
        await lock._acquire()
        lock.client._gets.assert_called_with(Keys.KEY)
        assert lock._token == lock.client._gets.return_value

    async def test_cas_calls_set_with_token(self, lock, mocker):
        m = mocker.spy(lock.client, "set")
        await lock._acquire()
        await lock.cas("value")
        m.assert_called_with(Keys.KEY, "value", _cas_token=lock._token)

    async def test_wrong_token_raises_error(self, mock_base_cache, lock):
        mock_base_cache._set.return_value = 0
        with pytest.raises(OptimisticLockError):
            await lock.cas("value")



import pickle
import random

import pytest
from marshmallow import Schema, fields, post_load

try:
    import ujson as json  # noqa: I900
except ImportError:
    import json  # type: ignore[no-redef]

from aiocache.serializers import (
    BaseSerializer,
    JsonSerializer,
    NullSerializer,
    PickleSerializer,
    StringSerializer,
)
from ..utils import Keys


class MyType:
    MY_CONSTANT = "CONSTANT"

    def __init__(self, r=None):
        self.r = r or random.randint(1, 10)

    def __eq__(self, obj):
        return self.__dict__ == obj.__dict__


class MyTypeSchema(Schema, BaseSerializer):
    r = fields.Integer()
    encoding = "utf-8"

    def dumps(self, *args, **kwargs):
        return super().dumps(*args, **kwargs)

    def loads(self, *args, **kwargs):
        return super().loads(*args, **kwargs)

    @post_load
    def build_my_type(self, data, **kwargs):
        return MyType(**data)

    class Meta:
        strict = True


class TestNullSerializer:
    TYPES = (1, 2.0, "hi", True, ["1", 1], {"key": "value"}, MyType())

    @pytest.mark.parametrize("obj", TYPES)
    async def test_set_get_types(self, memory_cache, obj):
        memory_cache.serializer = NullSerializer()
        assert await memory_cache.set(Keys.KEY, obj) is True
        assert await memory_cache.get(Keys.KEY) is obj

    @pytest.mark.parametrize("obj", TYPES)
    async def test_add_get_types(self, memory_cache, obj):
        memory_cache.serializer = NullSerializer()
        assert await memory_cache.add(Keys.KEY, obj) is True
        assert await memory_cache.get(Keys.KEY) is obj

    @pytest.mark.parametrize("obj", TYPES)
    async def test_multi_set_multi_get_types(self, memory_cache, obj):
        memory_cache.serializer = NullSerializer()
        assert await memory_cache.multi_set([(Keys.KEY, obj)]) is True
        assert (await memory_cache.multi_get([Keys.KEY]))[0] is obj


class TestStringSerializer:
    TYPES = (1, 2.0, "hi", True, ["1", 1], {"key": "value"}, MyType())

    @pytest.mark.parametrize("obj", TYPES)
    async def test_set_get_types(self, cache, obj):
        cache.serializer = StringSerializer()
        assert await cache.set(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == str(obj)

    @pytest.mark.parametrize("obj", TYPES)
    async def test_add_get_types(self, cache, obj):
        cache.serializer = StringSerializer()
        assert await cache.add(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == str(obj)

    @pytest.mark.parametrize("obj", TYPES)
    async def test_multi_set_multi_get_types(self, cache, obj):
        cache.serializer = StringSerializer()
        assert await cache.multi_set([(Keys.KEY, obj)]) is True
        assert await cache.multi_get([Keys.KEY]) == [str(obj)]


class TestJsonSerializer:
    TYPES = (1, 2.0, "hi", True, ["1", 1], {"key": "value"})

    @pytest.mark.parametrize("obj", TYPES)
    async def test_set_get_types(self, cache, obj):
        cache.serializer = JsonSerializer()
        assert await cache.set(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == json.loads(json.dumps(obj))

    @pytest.mark.parametrize("obj", TYPES)
    async def test_add_get_types(self, cache, obj):
        cache.serializer = JsonSerializer()
        assert await cache.add(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == json.loads(json.dumps(obj))

    @pytest.mark.parametrize("obj", TYPES)
    async def test_multi_set_multi_get_types(self, cache, obj):
        cache.serializer = JsonSerializer()
        assert await cache.multi_set([(Keys.KEY, obj)]) is True
        assert await cache.multi_get([Keys.KEY]) == [json.loads(json.dumps(obj))]


class TestPickleSerializer:
    TYPES = (1, 2.0, "hi", True, ["1", 1], {"key": "value"}, MyType())

    @pytest.mark.parametrize("obj", TYPES)
    async def test_set_get_types(self, cache, obj):
        cache.serializer = PickleSerializer()
        assert await cache.set(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == pickle.loads(pickle.dumps(obj))

    @pytest.mark.parametrize("obj", TYPES)
    async def test_add_get_types(self, cache, obj):
        cache.serializer = PickleSerializer()
        assert await cache.add(Keys.KEY, obj) is True
        assert await cache.get(Keys.KEY) == pickle.loads(pickle.dumps(obj))

    @pytest.mark.parametrize("obj", TYPES)
    async def test_multi_set_multi_get_types(self, cache, obj):
        cache.serializer = PickleSerializer()
        assert await cache.multi_set([(Keys.KEY, obj)]) is True
        assert await cache.multi_get([Keys.KEY]) == [pickle.loads(pickle.dumps(obj))]


class TestAltSerializers:
    async def test_get_set_alt_serializer_functions(self, cache):
        cache.serializer = StringSerializer()
        await cache.set(Keys.KEY, "value", dumps_fn=lambda _: "v4lu3")
        assert await cache.get(Keys.KEY) == "v4lu3"
        assert await cache.get(Keys.KEY, loads_fn=lambda _: "value") == "value"

    async def test_get_set_alt_serializer_class(self, cache):
        my_serializer = MyTypeSchema()
        my_obj = MyType()
        cache.serializer = my_serializer
        await cache.set(Keys.KEY, my_obj)
        assert await cache.get(Keys.KEY) == my_serializer.loads(my_serializer.dumps(my_obj))



import asyncio
import logging

from aiocache import Cache
from aiocache.lock import RedLock


logger = logging.getLogger(__name__)
cache = Cache(Cache.REDIS, endpoint='127.0.0.1', port=6379, namespace='main')


async def expensive_function():
    logger.warning('Expensive is being executed...')
    await asyncio.sleep(1)
    return 'result'


async def my_view():

    async with RedLock(cache, 'key', lease=2):  # Wait at most 2 seconds
        result = await cache.get('key')
        if result is not None:
            logger.info('Found the value in the cache hurray!')
            return result

        result = await expensive_function()
        await cache.set('key', result)
        return result


async def concurrent():
    await asyncio.gather(my_view(), my_view(), my_view())


async def test_redis():
    await concurrent()
    await cache.delete("key")
    await cache.close()


if __name__ == '__main__':
    asyncio.run(test_redis())



import pickle
from collections import namedtuple
from unittest import mock

import pytest

from aiocache.serializers import (
    BaseSerializer,
    JsonSerializer,
    MsgPackSerializer,
    NullSerializer,
    PickleSerializer,
    StringSerializer,
)


Dummy = namedtuple("Dummy", "a, b")

TYPES = [1, 2.0, "hi", True, ["1", 1], {"key": "value"}, Dummy(1, 2)]
JSON_TYPES = [1, 2.0, "hi", True, ["1", 1], {"key": "value"}]


class TestBaseSerializer:
    def test_init(self):
        serializer = BaseSerializer()
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "utf-8"

    def test_init_encoding(self):
        serializer = BaseSerializer(encoding="whatever")
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "whatever"

    def test_dumps(self):
        with pytest.raises(NotImplementedError):
            BaseSerializer().dumps("")

    def test_loads(self):
        with pytest.raises(NotImplementedError):
            BaseSerializer().loads("")


class TestNullSerializer:
    def test_init(self):
        serializer = NullSerializer()
        assert isinstance(serializer, BaseSerializer)
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "utf-8"

    @pytest.mark.parametrize("obj", TYPES)
    def test_set_types(self, obj):
        assert NullSerializer().dumps(obj) is obj

    def test_loads(self):
        assert NullSerializer().loads("hi") == "hi"


class TestStringSerializer:
    def test_init(self):
        serializer = StringSerializer()
        assert isinstance(serializer, BaseSerializer)
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "utf-8"

    @pytest.mark.parametrize("obj", TYPES)
    def test_set_types(self, obj):
        assert StringSerializer().dumps(obj) == str(obj)

    def test_loads(self):
        assert StringSerializer().loads("hi") == "hi"


class TestPickleSerializer:
    @pytest.fixture
    def serializer(self):
        yield PickleSerializer(protocol=4)

    def test_init(self, serializer):
        assert isinstance(serializer, PickleSerializer)
        assert serializer.DEFAULT_ENCODING is None
        assert serializer.encoding is None
        assert serializer.protocol == 4

    def test_init_sets_default_protocol(self):
        serializer = PickleSerializer()
        assert serializer.protocol == pickle.DEFAULT_PROTOCOL

    @pytest.mark.parametrize("obj", TYPES)
    def test_set_types(self, obj, serializer):
        assert serializer.loads(serializer.dumps(obj)) == obj

    def test_dumps(self, serializer):
        expected = b"\x80\x04\x95\x06\x00\x00\x00\x00\x00\x00\x00\x8c\x02hi\x94."
        assert serializer.dumps("hi") == expected

    def test_dumps_with_none(self, serializer):
        assert isinstance(serializer.dumps(None), bytes)

    def test_loads(self, serializer):
        assert serializer.loads(b"\x80\x03X\x02\x00\x00\x00hiq\x00.") == "hi"

    def test_loads_with_none(self, serializer):
        assert serializer.loads(None) is None

    def test_dumps_and_loads(self, serializer):
        obj = Dummy(1, 2)
        assert serializer.loads(serializer.dumps(obj)) == obj


class TestJsonSerializer:
    def test_init(self):
        serializer = JsonSerializer()
        assert isinstance(serializer, BaseSerializer)
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "utf-8"

    @pytest.mark.parametrize("obj", JSON_TYPES)
    def test_set_types(self, obj):
        serializer = JsonSerializer()
        assert serializer.loads(serializer.dumps(obj)) == obj

    def test_dumps(self):
        assert (
            JsonSerializer().dumps({"hi": 1}) == '{"hi": 1}'
            or JsonSerializer().dumps({"hi": 1}) == '{"hi":1}'  # json
        )  # ujson

    def test_dumps_with_none(self):
        assert JsonSerializer().dumps(None) == "null"

    def test_loads_with_null(self):
        assert JsonSerializer().loads("null") is None

    def test_loads_with_none(self):
        assert JsonSerializer().loads(None) is None

    def test_dumps_and_loads(self):
        obj = {"hi": 1}
        serializer = JsonSerializer()
        assert serializer.loads(serializer.dumps(obj)) == obj


class TestMsgPackSerializer:
    def test_init(self):
        serializer = MsgPackSerializer()
        assert isinstance(serializer, BaseSerializer)
        assert serializer.DEFAULT_ENCODING == "utf-8"
        assert serializer.encoding == "utf-8"

    def test_init_fails_if_msgpack_not_installed(self):
        with mock.patch("aiocache.serializers.serializers.msgpack", None):
            with pytest.raises(RuntimeError):
                MsgPackSerializer()
            assert JsonSerializer(), "Other serializers should still initialize"

    def test_init_use_list(self):
        serializer = MsgPackSerializer(use_list=True)
        assert serializer.use_list is True

    @pytest.mark.parametrize("obj", JSON_TYPES)
    def test_set_types(self, obj):
        serializer = MsgPackSerializer()
        assert serializer.loads(serializer.dumps(obj)) == obj

    def test_dumps(self):
        assert MsgPackSerializer().dumps("hi") == b"\xa2hi"

    def test_dumps_with_none(self):
        assert isinstance(MsgPackSerializer().dumps(None), bytes)

    def test_loads(self):
        assert MsgPackSerializer().loads(b"\xa2hi") == "hi"

    def test_loads_no_encoding(self):
        assert MsgPackSerializer(encoding=None).loads(b"\xa2hi") == b"hi"

    def test_loads_with_none(self):
        assert MsgPackSerializer().loads(None) is None

    def test_dumps_and_loads_tuple(self):
        serializer = MsgPackSerializer()
        assert serializer.loads(serializer.dumps(Dummy(1, 2))) == [1, 2]

    def test_dumps_and_loads_dict(self):
        serializer = MsgPackSerializer()
        d = {"a": [1, 2, ("1", 2)], "b": {"b": 1, "c": [1, 2]}}
        assert serializer.loads(serializer.dumps(d)) == {
            "a": [1, 2, ["1", 2]],
            "b": {"b": 1, "c": [1, 2]},
        }



import logging
import urllib
from copy import deepcopy
from typing import Dict

from aiocache import AIOCACHE_CACHES
from aiocache.base import BaseCache
from aiocache.exceptions import InvalidCacheType


logger = logging.getLogger(__name__)


def _class_from_string(class_path):
    class_name = class_path.split(".")[-1]
    module_name = class_path.rstrip(class_name).rstrip(".")
    return getattr(__import__(module_name, fromlist=[class_name]), class_name)


def _create_cache(cache, serializer=None, plugins=None, **kwargs):
    if serializer is not None:
        cls = serializer.pop("class")
        cls = _class_from_string(cls) if isinstance(cls, str) else cls
        serializer = cls(**serializer)

    plugins_instances = []
    if plugins is not None:
        for plugin in plugins:
            cls = plugin.pop("class")
            cls = _class_from_string(cls) if isinstance(cls, str) else cls
            plugins_instances.append(cls(**plugin))

    cache = _class_from_string(cache) if isinstance(cache, str) else cache
    instance = cache(serializer=serializer, plugins=plugins_instances, **kwargs)
    return instance


class Cache:
    """
    This class is just a proxy to the specific cache implementations like
    :class:`aiocache.SimpleMemoryCache`, :class:`aiocache.RedisCache` and
    :class:`aiocache.MemcachedCache`. It is the preferred method of
    instantiating new caches over using the backend specific classes.

    You can instatiate a new one using the ``cache_type`` attribute like:

    >>> from aiocache import Cache
    >>> Cache(Cache.REDIS)
    RedisCache (127.0.0.1:6379)

    If you don't specify anything, ``Cache.MEMORY`` is used.

    Only ``Cache.MEMORY``, ``Cache.REDIS`` and ``Cache.MEMCACHED`` types
    are allowed. If the type passed is invalid, it will raise a
    :class:`aiocache.exceptions.InvalidCacheType` exception.
    """

    MEMORY = AIOCACHE_CACHES["memory"]
    REDIS = AIOCACHE_CACHES.get("redis")
    MEMCACHED = AIOCACHE_CACHES.get("memcached")

    def __new__(cls, cache_class=MEMORY, **kwargs):
        if not issubclass(cache_class, BaseCache):
            raise InvalidCacheType(
                "Invalid cache type, you can only use {}".format(list(AIOCACHE_CACHES.keys()))
            )
        instance = cache_class.__new__(cache_class, **kwargs)
        instance.__init__(**kwargs)
        return instance

    @classmethod
    def _get_cache_class(cls, scheme):
        return AIOCACHE_CACHES[scheme]

    @classmethod
    def get_scheme_class(cls, scheme):
        try:
            return cls._get_cache_class(scheme)
        except KeyError as e:
            raise InvalidCacheType(
                "Invalid cache type, you can only use {}".format(list(AIOCACHE_CACHES.keys()))
            ) from e

    @classmethod
    def from_url(cls, url):
        """
        Given a resource uri, return an instance of that cache initialized with the given
        parameters. An example usage:

        >>> from aiocache import Cache
        >>> Cache.from_url('memory://')
        <aiocache.backends.memory.SimpleMemoryCache object at 0x1081dbb00>

        a more advanced usage using queryparams to configure the cache:

        >>> from aiocache import Cache
        >>> cache = Cache.from_url('redis://localhost:10/1?pool_max_size=1')
        >>> cache
        RedisCache (localhost:10)
        >>> cache.db
        1
        >>> cache.pool_max_size
        1

        :param url: string identifying the resource uri of the cache to connect to
        """
        parsed_url = urllib.parse.urlparse(url)
        kwargs = dict(urllib.parse.parse_qsl(parsed_url.query))
        cache_class = Cache.get_scheme_class(parsed_url.scheme)

        if parsed_url.path:
            kwargs.update(cache_class.parse_uri_path(parsed_url.path))

        if parsed_url.hostname:
            kwargs["endpoint"] = parsed_url.hostname

        if parsed_url.port:
            kwargs["port"] = parsed_url.port

        if parsed_url.password:
            kwargs["password"] = parsed_url.password

        return Cache(cache_class, **kwargs)


class CacheHandler:

    _config: Dict[str, Dict[str, object]] = {
        "default": {
            "cache": "aiocache.SimpleMemoryCache",
            "serializer": {"class": "aiocache.serializers.StringSerializer"},
        }
    }

    def __init__(self):
        self._caches = {}

    def add(self, alias: str, config: Dict[str, object]) -> None:
        """
        Add a cache to the current config. If the key already exists, it
        will overwrite it::

            >>> caches.add('default', {
                    'cache': "aiocache.SimpleMemoryCache",
                    'serializer': {
                        'class': "aiocache.serializers.StringSerializer"
                    }
                })

        :param alias: The alias for the cache
        :param config: Mapping containing the cache configuration
        """
        self._config[alias] = config

    def get(self, alias: str) -> object:
        """
        Retrieve cache identified by alias. Will return always the same instance

        If the cache was not instantiated yet, it will do it lazily the first time
        this is called.

        :param alias: str cache alias
        :return: cache instance
        """
        try:
            return self._caches[alias]
        except KeyError:
            pass

        config = self.get_alias_config(alias)
        cache = _create_cache(**deepcopy(config))
        self._caches[alias] = cache
        return cache

    def create(self, alias: str, **kwargs):
        """Create a new cache.

        You can use kwargs to pass extra parameters to configure the cache.

        :param alias: alias to pull configuration from
        :return: New cache instance
        """
        config = self.get_alias_config(alias)
        # TODO(PY39): **config | kwargs
        return _create_cache(**{**config, **kwargs})

    def get_alias_config(self, alias):
        config = self.get_config()
        if alias not in config:
            raise KeyError(
                "Could not find config for '{0}', ensure you include {0} when calling"
                "caches.set_config specifying the config for that cache".format(alias)
            )

        return config[alias]

    def get_config(self):
        """
        Return copy of current stored config
        """
        return deepcopy(self._config)

    def set_config(self, config):
        """
        Set (override) the default config for cache aliases from a dict-like structure.
        The structure is the following::

            {
                'default': {
                    'cache': "aiocache.SimpleMemoryCache",
                    'serializer': {
                        'class': "aiocache.serializers.StringSerializer"
                    }
                },
                'redis_alt': {
                    'cache': "aiocache.RedisCache",
                    'endpoint': "127.0.0.10",
                    'port': 6378,
                    'serializer': {
                        'class': "aiocache.serializers.PickleSerializer"
                    },
                    'plugins': [
                        {'class': "aiocache.plugins.HitMissRatioPlugin"},
                        {'class': "aiocache.plugins.TimingPlugin"}
                    ]
                }
            }

        'default' key must always exist when passing a new config. Default configuration
        is::

            {
                'default': {
                    'cache': "aiocache.SimpleMemoryCache",
                    'serializer': {
                        'class': "aiocache.serializers.StringSerializer"
                    }
                }
            }

        You can set your own classes there.
        The class params accept both str and class types.

        All keys in the config are optional, if they are not passed the defaults
        for the specified class will be used.

        If a config key already exists, it will be updated with the new values.
        """
        if "default" not in config:
            raise ValueError("default config must be provided")
        for config_name in config.keys():
            self._caches.pop(config_name, None)
        self._config = config


caches = CacheHandler()
